#!/usr/bin/python3

"""
Name: gsettings-datetime.py - [The GSettings Project](https://fox-echo.info/gsettings/)
Brief: clock (date and time) setup
Author: Frank ENDRES <frankendres.tuxfamily.org> (disabled; use <first@last.fr>)
Version: v0.6
License: CeCILL - http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html (GPL compatible)
"""

from typing import List, Optional
import os
import subprocess
import datetime
import time
import gettext; _ = gettext.gettext
import gi
import sys



# GTK3 / 4 compatibility layer -> code is written in GTK4 style
if sys.argv[0].find('-gtk3.') != -1: #GTK3
    gi.require_version('Gtk', '3.0'); from gi.repository import GLib, Gtk, Gio
    class CompatApplicationWindow(Gtk.ApplicationWindow): #GTK3
        def set_child(self, child: Gtk.Widget): #renamed method
            for content in self.get_children(): #keep header bar and remove other content
                if not isinstance(content, CompatHeaderBar):
                    self.remove(content)
            self.add(child)
        def present(self): #renamed method
            self.show_all()
            super().present()
        def connect(self, event: str, callback):
            if event == 'close-request':
                def close_callback(window, event): #changed event name and callback profile
                    callback(window)
                super().connect('delete-event', close_callback)
            else:
                super().connect(event, callback)
    class CompatWindow(Gtk.Window):
        def set_child(self, child: Gtk.Widget): #renamed method
            self.add(child)
        def present(self): #renamed method
            self.show_all()
            super().present()
        def set_default_widget(self, widget: Gtk.Widget):
            widget.set_can_default(True)
            self.set_default(widget)
    class CompatScrolledWindow(Gtk.ScrolledWindow):
        def set_child(self, child: Gtk.Widget): #renamed method
            self.add(child)
    class CompatBox(Gtk.Box):
        def append(self, widget: Gtk.Widget): #renamed method
            self.pack_start(widget, False, False, 0)
    class CompatButton(Gtk.Button):
        def new_from_icon_name(icon_name: str): #changed method profile
            #Gtk.Button.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON) returns a Gtk.Button not a CompatButton
            button = CompatButton()
            button.set_icon_name(icon_name)
            return button
        def set_icon_name(self, icon_name: str):
            image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
            self.set_image(image)
    class CompatRadioButton(Gtk.RadioButton):
        def set_group(self, radio): #renamed method
            self.join_group(radio)
    class CompatAboutDialog(Gtk.AboutDialog):
        def set_visible(self, visible: bool): #renamed method
            if visible == True:
                super().run()
                self.destroy()
    class CompatHeaderBar(Gtk.HeaderBar):
        def __init__(self):
            super().__init__()
            self.set_show_close_button(True)
        def set_title_widget(self, label: Gtk.Label):
            self.props.title = label.get_label()
    class CompatPasswordEntry(Gtk.Box):
        def __init__(self):
            super().__init__(orientation = Gtk.Orientation.HORIZONTAL, spacing=10)
            self.password_entry = Gtk.Entry()
            self.password_entry.set_visibility(False)
            self.password_entry.set_activates_default(True)
            self.pack_start(self.password_entry, True, True, 0)
            show_password_button = CompatButton.new_from_icon_name('media-optical-symbolic')
            show_password_button.set_tooltip_text(_('show / hide password'))
            self.pack_start(show_password_button, False, False, 0)
            show_password_button.connect('clicked', lambda bt: self.password_entry.set_visibility(not self.password_entry.get_visibility()))
        def get_text(self) -> str:
            return self.password_entry.get_text()
        def connect(self, event, callback):
            self.password_entry.connect(event, callback)
    class CompatStringListStore(Gtk.ListStore):
        def __init__(self):
            super().__init__(str)
        def append(self, value: str):
            super().append([value])
        def remove_all(self):
            self.clear()
    class CompatDropDown(Gtk.Entry):
        def __init__(self):
            super().__init__()
            self.completion = Gtk.EntryCompletion()
            self.completion.set_text_column(0)
            self.completion.set_match_func(lambda completion, key, itr: key.lower() in self.completion.get_model()[itr][0].lower())
            self.set_completion(self.completion)
        def connect(self, event, callback):
            super().connect("activate", callback, None) #notify::selected
        def get_model(self) -> CompatStringListStore:
            return self.completion.get_model()
        def set_model(self, model: CompatStringListStore):
            self.completion.set_model(model)
        def get_selected_item(self) -> Optional[str]:
            selected = None
            iter = self.completion.get_model().get_iter_first()
            while selected == None and iter is not None:
                if self.get_text() == self.completion.get_model().get_value(iter, 0):
                    selected = self.get_text()
                iter = self.completion.get_model().iter_next(iter)
            return selected
        def set_selected_item(self, value: str):
            self.set_text(value)
    class CompatCalendar(Gtk.Calendar):
        def select_day(self, gDate):
            self.select_month(gDate.get_month() - 1, gDate.get_year())
            super().select_day(gDate.get_day_of_month())
else: #GTK4
    gi.require_version('Gtk', '4.0'); from gi.repository import GLib, Gtk, Gio
    class CompatApplicationWindow(Gtk.ApplicationWindow):
        pass
    class CompatWindow(Gtk.Window):
        pass
    class CompatScrolledWindow(Gtk.ScrolledWindow):
        pass
    class CompatBox(Gtk.Box):
        pass
    class CompatButton(Gtk.Button):
        def new_from_icon_name(icon_name: str):
            return Gtk.Button.new_from_icon_name(icon_name)
    class CompatRadioButton(Gtk.CheckButton):
        pass
    class CompatAboutDialog(Gtk.AboutDialog):
        pass
    class CompatHeaderBar(Gtk.HeaderBar):
        pass
    class CompatPasswordEntry(Gtk.PasswordEntry):
        def __init__(self):
            super().__init__()
            self.set_show_peek_icon(True)
    class CompatStringListStore(Gio.ListStore):
        #Gio.ListStore.new(Gtk.StringObject)
        def append(self, value: str):
            super().append(Gtk.StringObject.new(value))
    class CompatDropDown(Gtk.DropDown):
        def __init__(self):
            super().__init__()
            self.set_enable_search(True)
            self.set_search_match_mode(Gtk.StringFilterMatchMode.SUBSTRING)
            self.set_expression(Gtk.PropertyExpression.new(Gtk.StringObject, None, 'string'))
        def get_selected_item(self) -> Optional[str]: #added method
            item = super().get_selected_item()
            return None if item == None else item.get_string()
        def set_selected_item(self, value: str): #added method
            found, idx = self.get_model().find_with_equal_func(Gtk.StringObject.new(value), lambda item, value, _: item.get_string() == value.get_string(), None)
            self.set_selected(idx if found else Gtk.INVALID_LIST_POSITION)
    class CompatCalendar(Gtk.Calendar):
        pass



class DateTimeSettings(Gtk.Application):
    def __init__(self):
        super().__init__(application_id='org.gsettings.datetime') #flags=Gio.ApplicationFlags.DEFAULT_FLAGS


    def do_activate(self):
        self.win = self.props.active_window
        if self.win == None:
            self.win = CompatApplicationWindow(application=self)
            self.win.set_default_icon_name('gsettings-datetime')
            header_bar = CompatHeaderBar()
            header_bar.set_title_widget(Gtk.Label(label=_('Clock Configuration')))
            about_button = CompatButton.new_from_icon_name('help-about')
            about_button.connect('clicked', self.show_about_dialog)
            header_bar.pack_start(about_button)
            self.win.set_titlebar(header_bar)
            self.win.set_resizable(False)
            self.grid = Gtk.Grid()
            self.grid.set_margin_start(10); self.grid.set_margin_end(10); self.grid.set_margin_top(10); self.grid.set_margin_bottom(10)
            self.grid.set_row_spacing(10)
            self.win.set_child(self.grid)
            self.add_widgets()

        self.win.present()
        self.on_reload_clicked(None)


    def add_widgets(self):
        # NTP
        self.ntp_button = Gtk.CheckButton(label=_('NTP sync'))
        self.ntp_button.set_tooltip_text(_('Enable / disable time synchronization.'))
        self.grid.attach(self.ntp_button, 1, 0, 1, 1)
        self.ntp_button.connect('toggled', self.on_ntp_clicked)

        # Date
        self.date_label = Gtk.Label(label=_('Date:'))
        self.grid.attach(self.date_label, 0, 1, 1, 1)
        self.calendar = CompatCalendar()
        self.grid.attach(self.calendar, 1, 1, 1, 1)

        # Time
        self.time_label = Gtk.Label(label=_('Time:'))
        self.grid.attach(self.time_label, 0, 2, 1, 1)
        self.time_box = CompatBox(orientation = Gtk.Orientation.HORIZONTAL)
        self.grid.attach(self.time_box, 1, 2, 1, 1)
        self.hour_spin = Gtk.SpinButton(orientation = Gtk.Orientation.VERTICAL, adjustment = Gtk.Adjustment(lower = 0, upper = 23, step_increment = 1), numeric = True, wrap = True)
        self.time_box.append(self.hour_spin)
        self.time_box.append(Gtk.Label(label = '<span weight="bold"> : </span>', use_markup = True))
        self.minutes_spin = Gtk.SpinButton(orientation = Gtk.Orientation.VERTICAL, adjustment = Gtk.Adjustment(lower = 0, upper = 59, step_increment = 1), numeric = True, wrap = True)
        self.time_box.append(self.minutes_spin)

        # Timezone
        timezone_label = Gtk.Label(label=_('Timezone:'))
        self.grid.attach(timezone_label, 0, 3, 1, 1)
        self.tz_entry = CompatDropDown()
        self.tz_entry.set_model(CompatStringListStore())
        self.grid.attach(self.tz_entry, 1, 3, 1, 1)
        for tz in load_tz():
            self.tz_entry.get_model().append(tz)

        # System time (UTC)
        st_label = Gtk.Label(label=_('System clock:'))
        self.grid.attach(st_label, 0, 4, 1, 1)
        self.utc_button = CompatRadioButton()
        self.utc_button.set_label(_('UTC'))
        self.utc_button.set_tooltip_text(_('Set hardware clock to UTC (Linux default).'))
        self.local_button = CompatRadioButton()
        self.local_button.set_label(_('localtime'))
        self.local_button.set_tooltip_text(_('Set hardware clock to local time (Windows default).'))
        self.local_button.set_group(self.utc_button)
        st_vbox = CompatBox(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        st_vbox.append(self.utc_button)
        st_vbox.append(self.local_button)
        self.grid.attach(st_vbox, 1, 4, 1, 1)

        # Buttons
        vbox = CompatBox(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
        self.grid.attach(vbox, 1, 5, 1, 1)
        reload_button = Gtk.Button(label=_('Reload'))
        reload_button.connect('clicked', self.on_reload_clicked)
        vbox.append(reload_button)
        save_button = Gtk.Button(label=_('Save'))
        save_button.connect('clicked', self.on_save_clicked)
        vbox.append(save_button)


    def on_reload_clicked(self, button):
        self.ntp_button.set_active(get_ntp_status())
        year, month, day, hour, minutes, sec, wday, yday, isdst = datetime.datetime.now().timetuple()
        self.calendar.select_day(GLib.DateTime.new_local(year, month, day, hour, minutes, sec))
        self.hour_spin.set_value(hour)
        self.minutes_spin.set_value(minutes)
        self.tz_entry.set_selected_item(get_tz())
        self.utc_button.set_active(get_hw_hour() == time.gmtime().tm_hour)


    def show_about_dialog(self, widget):
        about_dialog = CompatAboutDialog(transient_for=self.win, modal=True)
        about_dialog.set_program_name('GSettings-DateTime')
        about_dialog.set_version('v0.6')
        about_dialog.set_authors(['Frank ENDRES <first@last.fr>'])
        about_dialog.set_translator_credits('\n- de: Frank ENDRES\n- fr: Frank ENDRES')
        about_dialog.set_website('https://fox-echo.info/gsettings/')
        about_dialog.set_license('- CeCILL (Ce[a] C[nrs] I[nria] L[ogiciel] L[ibre])\n- GPL (General Public License)')
        about_dialog.set_logo_icon_name('gsettings-datetime')
        about_dialog.set_visible(True)


    def on_ntp_clicked(self, button):
        if self.ntp_button.get_active():
            self.date_label.set_visible(False)
            date = self.calendar.set_visible(False)
            self.time_label.set_visible(False)
            time = self.time_box.set_visible(False)
        else:
            self.date_label.set_visible(True)
            date = self.calendar.set_visible(True)
            self.time_label.set_visible(True)
            time = self.time_box.set_visible(True)


    def on_save_clicked(self, button):
        ntp = self.ntp_button.get_active()
        set_ntp(ntp)
        tz = self.tz_entry.get_selected_item()
        print(tz); return #DEBUG
        utc = self.utc_button.get_active()
        if tz != None:
            set_tz(tz)
            if not ntp:
                year, month, day = self.calendar.get_date()
                date = datetime.datetime(year, month + 1, day).strftime("%Y-%m-%d")
                time = datetime.datetime(2001, 1, 1, int(self.hour_spin.get_value()), int(self.minutes_spin.get_value())).strftime("%H:%M")
                set_datetime(date, time)
                set_hwclock(date, time, utc)
            else:
                set_hwclock(datetime.datetime.now().strftime('%Y-%m-%d'), datetime.datetime.now().strftime('%H:%M'), utc)
            app.quit()





# Minimalistic TimeZone library (no PyTZ)
ZONE_INFO = '/usr/share/zoneinfo'

def get_tz() -> str:
    tz = None
    if os.path.exists('/etc/localtime-copied-from'): #Slackware
        tz = os.path.realpath('/etc/localtime-copied-from')[len(ZONE_INFO) + 1:]
    elif os.path.exists('/etc/localtime'):
        tz = os.path.realpath('/etc/localtime')[len(ZONE_INFO) + 1:]
    return tz

def load_tz() -> List[str]:
    timezones = []
    _load_tz(ZONE_INFO, timezones)
    return timezones

def _load_tz(folder: str, timezones : List[str]):
    for entry in os.listdir(folder):
        path = os.path.join(folder, entry)
        if not os.path.isdir(path):
            timezones.append(path[len(ZONE_INFO) + 1:])
        elif path != '/usr/share/zoneinfo/right' and path != '/usr/share/zoneinfo/posix':
            _load_tz(path, timezones)

def set_tz(tz: str):
    if tz != None:
        if os.path.exists('/etc/localtime-copied-from'): #Slackware
            subprocess.run(['cp', '-f', ZONE_INFO + '/' + tz, '/etc/localtime'])
            subprocess.run(['ln', '-sf', ZONE_INFO + '/' + tz, '/etc/localtime-copied-from'])
        else:
            subprocess.run(['ln', '-sf', ZONE_INFO + '/' + tz, '/etc/localtime'])





# Minimalistic time management wrapper
def get_hw_hour() -> int:
    f = open('/sys/class/rtc/rtc0/time', 'r')
    hw_hour = int(f.read()[:2])
    f.close()
    return hw_hour

def set_datetime(date: str, time: str):
    subprocess.run(['date', '-s', date + ' ' + time])

def set_hwclock(date: str, time: str, utc: bool):
    if os.path.exists('/etc/slackware-version'): #Slackware
        f = open('/etc/hardwareclock', 'w')
        f.write('UTC' if utc else 'localtime')
        f.close()
    subprocess.run(['hwclock', '--set', '--date', date + ' ' + time, '--utc' if utc else '--localtime'])





# Minimalistic NTP wrapper
def get_ntp_status() -> bool:
    if os.path.exists('/etc/rc.d/rc.ntpd'): #Slackware
        ntp = os.access('/etc/rc.d/rc.ntpd', os.X_OK)
    elif os.path.exists('/usr/lib/systemd/system/systemd-timesyncd.service'): #SystemD
        ntp = subprocess.run(['systemctl', 'status', 'systemd-timesyncd'], check=False, capture_output=True).returncode == 0
    #elif: #TODO: sysvinit, runit, ...
    else:
        return None
    return ntp

def set_ntp(status: bool):
    if (status != get_ntp_status()):
        if os.path.exists('/etc/rc.d/rc.ntpd'): #Slackware
            if status:
                subprocess.run(['chmod', '+x', '/etc/rc.d/rc.ntpd'])
                subprocess.run(['/etc/rc.d/rc.ntpd', 'start'])
            else:
                subprocess.run(['/etc/rc.d/rc.ntpd', 'stop'])
                subprocess.run(['chmod', '-x', '/etc/rc.d/rc.ntpd'])
        elif os.path.exists('/usr/lib/systemd/system/systemd-timesyncd.service'): #SystemD
            subprocess.run(['systemctl', 'start' if status else 'stop', 'systemd-timesyncd'])
            subprocess.run(['systemctl', 'enable' if status else 'disable', 'systemd-timesyncd'])
        #elif: #TODO: sysvinit, runit, ...





if __name__ == '__main__':
    gettext.textdomain('gsettings-datetime')
    app = DateTimeSettings()
    app.run()
